-- EThreads (using pthread) by Jeffrey Fielding (JJProg@cyberbury.net)
-- Version 0.31 alpha
-- EUPHORIA DOES NOT SUPPORT REENTRANCY. USE AT YOUR OWN RISK.
-- For some reason, the join call doesn't wait for the process to finish
-- I call it anyway to try to free any resources
-- Then I have a seperate function that is always called first (before
-- the routine given to create_thread.) It then calls the appropriate
-- routine, and sets a status flag so you can use join
-- to wait for a thread to finish (it polls the flag).
-- Since I'm not sure how Euphoria might react to having threads accessing
-- variables at the same time, I use memory instead.
-- Therefore, any arguments to your thread must be in memory.
-- You may use the experimental create_thread2 procedure with an Euphoria
-- object instead, but I don't know how it will react.
-- id = create_thread(routine, argument)
-- returns -1 on an error. argument is a memory address.
-- You are responsible for freeing the argument after the thread is joined
-- id = create_thread2(routine, argument)
-- returns -1 on an error. argument is an Euphoria object.
-- join_thread(id)
-- waits for thread with id id to finish, and attempts to free any resources
-- used by it.
-- join_all_threads()
-- waits for all threads to finish and attemts to free all resources
include dll.e
include machine.e
constant PTHREAD_LIB = open_dll("/usr/lib/libpthread.so")
constant MYTHREAD_LIB = open_dll("./ethread.so") -- Added in version 0.3:
-- No way to call C routines with structures as arguments, so I added this
-- library I wrote in c to fix it.
constant CREATE = define_c_func(PTHREAD_LIB,"pthread_create",{C_POINTER,C_POINTER,C_POINTER,C_POINTER},C_INT)
--constant JOIN = define_c_func(PTHREAD_LIB,"pthread_join",{C_POINTER,C_POINTER},C_INT)
-- Join doesn't work because the first argument must be a structure.
-- Size is 4, so try a long.
--constant JOIN = define_c_func(PTHREAD_LIB,"pthread_join",{C_LONG,C_POINTER},C_INT)
constant JOIN = define_c_func(MYTHREAD_LIB,"fixed_join",{C_POINTER,C_POINTER},C_INT)
sequence threads, routines, args
threads = {}
routines = {}
args = {}
function callback1(atom x)
    call_proc(peek4u(x),{peek4u(x+4)})
    poke(x+8,1)
    return 0
end function
constant CALLBACK1 = call_back(routine_id("callback1"))
function callback2(atom x)
    call_proc(peek4u(x),{args[peek4u(x+4)]})
    poke(x+8,1)
    return 0
end function
constant CALLBACK2 = call_back(routine_id("callback2"))
global function create_thread(integer routine, atom arg)
    integer x, index
    index = find(-1,threads)
    if index = 0 then
	threads &= allocate(4)
	routines &= allocate(9)
	args &= 0
	index = length(threads)
    else
	threads[index] = allocate(8)
	routines[index] = allocate(9)
    end if
    poke(routines[index]+8,0)
    poke4(routines[index],{routine,arg})
    x = c_func(CREATE,{threads[index],0,CALLBACK1,routines[index]})
    if x != 0 then
	return -1
    end if
    return index
end function
global function create_thread2(integer routine, object arg)
    integer x, index
    index = find(-1,threads)
    if index = 0 then
	threads &= allocate(4)
	routines &= allocate(9)
	args = append(args,arg)
	index = length(threads)
    else
	args[index] = arg
	threads[index] = allocate(4)
	routines[index] = allocate(9)
    end if
    poke(routines[index]+4,0)
    poke4(routines[index],{routine,index})
    x = c_func(CREATE,{threads[index],0,CALLBACK2,routines[index]})
    if x != 0 then
	return -1
    end if
    return index
end function
global procedure join_thread(integer id)
    integer x
    atom r
    r = allocate(4)
    --x = c_func(JOIN,{threads[id],r})
    --printf(1,"Status before join: %d\n",peek(routines[id]+8))
    --x = c_func(JOIN,{peek4s(threads[id]),r})
    x = c_func(JOIN,{threads[id],r})
    ? x
    --printf(1,"Status after join: %d\n",peek(routines[id]+8))
    free(r)
    while peek(routines[id]+8) != 1 do
	--printf(1,"Waiting... %d for %d\n",{peek(routines[id]+8),id})
    end while
    free(threads[id])
    free(routines[id])
    threads[id] = -1
end procedure
global procedure join_all_threads()
    for i = 1 to length(threads) do
	if threads[i] != -1 then
	    join_thread(i)
	end if
    end for
end procedure
